local Sol = LibStub("Sol")

local BUTTON_SIZE = 16           -- the size of each button in the item list
local MIN_SIZE = {                 -- minimum size for the IP frame
    width = 320,
    height = 440
}

local MAX_SIZE = {		-- maximum size for the IP frame
    width = 1366,
    height = 768
}
local buttonsToShow = ItemPreview.constants.MAX_BUTTONS  -- the number of item buttons that can be shown (depends on the size of the frame)
local searchTime = nil             -- when the next search should be executed (set by QueueSearch when a user types)
local resizeTime = nil             -- time of the next resize update

-------------------------------------------------------------------------------
--  Interaction Functions                                                    --
-------------------------------------------------------------------------------
do

-- Updates the number of item buttons visible
function ItemPreview:CheckListBounds()
    -- ensure minimum size
    local w, h = self.frames.frmMain:GetSize()
    local update = false
    if w < MIN_SIZE.width then
        update = true
        w = MIN_SIZE.width
    end
    if h < MIN_SIZE.height then
        update = true
        h = MIN_SIZE.height
    end
    if h > MAX_SIZE.height then
        update = true
        h = MAX_SIZE.height
    end
    if update then
        self.frames.frmMain:SetSize(w, h)
    end
    
    w, h = self.frames.lstItems:GetSize()
    -- show or hide list buttons as needed
    local numButtons = math.floor( (h - (BUTTON_SIZE/2)) / BUTTON_SIZE )
    if numButtons > self.constants.MAX_BUTTONS then
        numButtons = self.constants.MAX_BUTTONS 
    end
    if numButtons < self.constants.MIN_BUTTONS then
        numButtons = self.constants.MIN_BUTTONS
    end
    if numButtons ~= buttonsToShow then
        for i = 1, numButtons do
            _G[self.frames.lstItems:GetName() .. "Item" .. i]:Show()
        end
        for i = numButtons + 1, self.constants.MAX_BUTTONS do
            _G[self.frames.lstItems:GetName() .. "Item" .. i]:Hide()
        end

        local startID = math.max(self.frames.txtStartID:GetNumber(), self.itemList.minID)
        local endID = math.min(self.frames.txtEndID:GetNumber(), self.itemList.maxID)
        local quality = UIDropDownMenu_GetSelectedID(self.frames.ddQuality)
        local matchCount = 0   -- how many items match search criteria so far
        for i = 1, #self.sortedList do
            local id = self.sortedList[i]
            local item = nil
            if id >= startID and id <= endID then
                item = self.itemList[id]
            end
            if item and self:ShouldShowItem(item, quality) then
                matchCount = matchCount + 1
            end
        end
        
        local scrollBar = self.frames.sbItems
        local maxVal = matchCount - buttonsToShow
        if maxVal < 0 then
            maxVal = 0
        end
        scrollBar:SetMinMaxValues(0, maxVal)
	
        UIPanelScrollBar_OnValueChanged(scrollBar)
    end
    buttonsToShow = numButtons
end

-- Shows a tooltip for the item link
function ItemPreview:ItemButton_OnEnter(button)
    if button.label:GetText() == "" then return end
    
    GameTooltip:SetOwner(button, "ANCHOR_BOTTOMRIGHT")
    GameTooltip:SetHyperLink(button.label:GetText())
    GameTooltip1:Hide()
    GameTooltip2:Hide()

    if not button.itemID then return end
    local item = self.itemList[button.itemID]
    
    if self.settings.ShowZone then
        local convertFn = function(zoneID)
            return GetZoneLocalName(zoneID)
        end
        local data = self:GetItemFieldString(item, "zones", convertFn)
        if data then
            GameTooltip:AddLine(" ")
            GameTooltip:AddLine("Seen in: " .. data)
        end
    end
    
    if self.settings.ShowMobs then
        local data = self:GetItemFieldString(item, "mobs")
        if data then
            GameTooltip:AddLine(" ")
            GameTooltip:AddLine("Dropped by: " .. data)
        end
    end
    
    GameTooltip:AddLine(" ")
    if self.settings.ShowUpdateTime then
        local date = item.date or self.itemList.originalRefreshDate
        if date then
            GameTooltip:AddLine("Updated on " .. date)
        end
    end
    GameTooltip:AddLine("Item ID: " .. button.itemID)
end

-- Start resizing the list and any subcomponents that should be resized with it
function ItemPreview:StartSizingList()
    self.frames.frmMain:StartSizing("BOTTOMRIGHT")
    resizeTime = GetTime() + self.settings.ResizeTime
end

-- Stop resizing the list and any subcomponents that should be resized with it
function ItemPreview:StopSizingList()
    Sol.config.SaveFrameBounds(self.name, self.frames.frmMain) 
    resizeTime = nil
    self:CheckListBounds()
end

-- Creates a concatenated string of the values in the specified field of a saved item
function ItemPreview:GetItemFieldString(item, field, convert)
    if not item or not item[field] then return nil end
    
    local invertedList = {}
    local count = 1
    for k, v in pairs(item[field]) do
        invertedList[count] = (convert and convert(k)) or k
        invertedList[count] = invertedList[count] .. "(" .. v .. ")"
        count = count + 1
    end
    return table.concat(invertedList, ", ")
end

-- Shows visiual cues for a selected item
function ItemPreview:SelectItem(button, key)
    if not button.itemID then return end
    
    local list = button:GetParent()
    if list.selectedButton then
        list.selectedButton:UnlockHighlight()
    end
    button:LockHighlight()
    list.selectedButton = button
    list.selectedID = button.itemID
    
    local link = button.label:GetText()
    self:ShowLink(link)
    GameTooltip:Hide()
end

function ItemPreview:Sort(sortFn)
    local sortedTable = {}
    for k, v in pairs(self.itemList) do
        if type(k) == "number" then
            table.insert(sortedTable, k)
        end
    end
    table.sort(sortedTable, sortFn)
    self.sortedList = sortedTable
    self:UpdateList(true)
end

end -- interaction functions


-------------------------------------------------------------------------------
--  List Update Functions                                                    --
-------------------------------------------------------------------------------
do

-- Runs through the items list, updates the scroll bar, and shows those items 
-- that should be shown based on the id range and search criteria
function ItemPreview:UpdateList(isScrolling)    
    if not self.sortedList then return end
    
    -- If a search has been queued, ignore the scroll and do the search instead
    if searchTime then 
        isScrolling = false 
    end
    
    local startID = math.max(self.frames.txtStartID:GetNumber(), self.itemList.minID)
    local endID = math.min(self.frames.txtEndID:GetNumber(), self.itemList.maxID)
    
    local offset = self.frames.sbItems:GetValue()
    if offset < 0 then
        offset = 0
    end
    
    for i = 1, self.constants.MAX_BUTTONS do
        local button = _G[self.frames.lstItems:GetName() .. "Item" .. i]
        button:Hide()
    end
    
    -- Clear visible selection
    if self.frames.lstItems.selectedButton then
        self.frames.lstItems.selectedButton:UnlockHighlight()
        self.frames.lstItems.selectedButton = nil
    end
    
    local quality = UIDropDownMenu_GetSelectedID(self.frames.ddQuality)
    local buttonCount = 0  -- which physical button in the list we're on
    local matchCount = 0   -- how many items match search criteria so far
    for i = 1, #self.sortedList do
        local id = self.sortedList[i]
        local item = nil
        if id >= startID and id <= endID then
            item = self.itemList[id]
        end
        if item and ((isScrolling and item.show) or (not isScrolling and self:ShouldShowItem(item, quality))) then
            item.show = true
            -- item matches search criteria, but may not be in scroll range
            matchCount = matchCount + 1
            if matchCount > offset and buttonCount < buttonsToShow then 
                -- item is in scroll range
                buttonCount = buttonCount + 1
                local button = _G[self.frames.lstItems:GetName() .. "Item" .. buttonCount]
                button:Show()
                button.itemID = id
                button.label:SetText(self:MakeItemLink(string.format("%x", item.id), item.text, item.color))
                if self.frames.lstItems.selectedID == id then
                    button:LockHighlight()
                    self.frames.lstItems.selectedButton = button
                end
            end
            -- no need to go through the rest if we're just scrolling
            if isScrolling and buttonCount >= buttonsToShow then
                break
            end
        elseif item then
            item.show = nil
        end
    end
    
    if not self.frames.lstItems.selectedButton then
        self.frames.lstItems.selectedID = -1
    end
    
    if not isScrolling then
        local maxVal = matchCount - buttonsToShow
        if maxVal < 0 then
            maxVal = 0
        end
        self.frames.sbItems:SetMinMaxValues(0, maxVal)
        self.frames.lblTitle:SetText("ItemPreview (" .. matchCount .. " items)")
    end
    UIPanelScrollBar_OnValueChanged(self.frames.sbItems)
    searchTime = nil
end

-- Updates the scroll bar
function ItemPreview:Scroll(list, delta)
    local scrollBar = self.frames.sbItems
    local newVal = scrollBar:GetValue() + ((delta < 0 and (buttonsToShow - 1)) or (-1 * buttonsToShow + 1))
    if newVal >= scrollBar:GetMinValue() and newVal <= scrollBar:GetMaxValue() then
        scrollBar:SetValue(newVal)
    end
end

-- Called when the scroll bar value changes
function ItemPreview:OnScroll(item, delta)
    self:UpdateList(true)
end

-- Setup a search to be run when the user stops entering input
function ItemPreview:QueueSearch()
    searchTime = GetTime() + self.settings.SearchDelay
end

-- Process refresh, queued searches, and resize
function ItemPreview:ItemList_OnUpdate()
    if searchTime and GetTime() >= searchTime then
        self:AdvancedSearch()
        return
    end
    
    if resizeTime and GetTime() >= resizeTime then
        self:CheckListBounds()
        return
    end
    
    if not self.refreshState or not self.refreshState.isRefreshing then return end
    
    self:RunRefreshLoop()
    
    GameTooltip1:Hide()
    GameTooltip2:Hide()
end

end -- list update functions