local Sol = LibStub("Sol")

searchFields = nil    -- a table of fnKey->searchString where fnKey are the search criteria 

-------------------------------------------------------------------------------
--  Search Functions                                                         --
-------------------------------------------------------------------------------

-- Set up the SearchFunctions table, linking search keys to functions
function ItemPreview:InitSearchFunctions()
    self.SearchFunctions = {
        id = self.SearchForID,
        tip = self.SearchInTooltip,
        name = self.SearchForName,
        level = self.SearchForLevel,
        lvl = self.SearchForLevel,
        slot = self.SearchForEquipSlot,
        type = self.SearchForEquipType,
    }
end

-- Check if the given item goes in the slot specified by searchStr
function ItemPreview:SearchForEquipSlot(searchStr, item, ignoreCase)
    if not item.tip then return false end
    
    for k, v in pairs(item.tip.right) do
        if not v then return false end
        if ignoreCase then 
            v = v:lower()
        end
        if v == searchStr then
            return true
        end
    end
    return false
end

-- Check if the given item is of the type specified by searchStr
function ItemPreview:SearchForEquipType(searchStr, item, ignoreCase)
    if not item.tip then return false end

    for i = 2, 5 do
        if item.tip.left[i] then 
            local found = false
            if ignoreCase then 
                found = (item.tip.left[i]:lower() == searchStr)
            else
                found = (item.tip.left[i] == searchStr)
            end
            if found then return true end
        end
    end
    return false
end

-- Check if the specified item has a level and if it's within the parameters given in searchStr
-- searchStr should be of the form '#', '>#', '<#', or '>#<#'
function ItemPreview:SearchForLevel(searchStr, item, ignoreCase)    
    if not item.lvl then return false end
    
    local minLevel, maxLevel = self:GetNumberRange(searchStr)
    return item.lvl >= minLevel and item.lvl <= maxLevel
end

-- Check if the specified item's id is within the parameters given in searchStr
-- searchStr should be of the form '#', '>#', '<#', or '>#<#'
function ItemPreview:SearchForID(searchStr, item, ignoreCase)    
    local minID, maxID = self:GetNumberRange(searchStr)
    return item.id >= minID and item.id <= maxID
end

-- Parse a min and max value out of a string in the form '#', '>#', '<#', or '>#<#'
function ItemPreview:GetNumberRange(searchStr)
    local min, max
    local firstChar = searchStr:sub(1, 1)
    if firstChar == "<" then
        local gtIndex = searchStr:find(">")
        if gtIndex and gtIndex > 0 then
            min = tonumber(searchStr:sub(gtIndex + 1))
            searchStr = searchStr:sub(0, gtIndex - 1)
        else
            min = 0
        end
        max = tonumber(searchStr:sub(2))
    elseif firstChar == ">" then
        local ltIndex = searchStr:find("<")
        if ltIndex and ltIndex > 0 then
            max = tonumber(searchStr:sub(ltIndex + 1))
            searchStr = searchStr:sub(0, ltIndex - 1)
        else
            max = 99999999
        end
        min = tonumber(searchStr:sub(2))
    else
        min = tonumber(searchStr)
        max = tonumber(searchStr)
    end
    return min, max
end

-- Search for the specified string in the name (tooltip first line) of the given item
function ItemPreview:SearchForName(searchStr, item, ignoreCase)    
    if ignoreCase then 
        return item.text:lower():match(searchStr)
    else
        return item.text:match(searchStr)
    end
end

-- Search for the specified string in the tooltip of the given item
function ItemPreview:SearchInTooltip(searchStr, item, ignoreCase)
    if not item.tip then return false end
    for i = 1, self.constants.TOOLTIP_LINES do
        local left = item.tip.left[i] or ""
        local right = item.tip.right[i] or ""
        if ignoreCase then
            left = left:lower()
            right = right:lower()
        end
        if left:match(searchStr) or right:match(searchStr) then
            return true
        end
    end
    return false
end

-- Run a search using all the fields in the advanced search window
function ItemPreview:AdvancedSearch()
    local fields = {}
   
    do -- level
    local minLevel = self.frames.txtMinLevel:GetText() or ""
    local maxLevel = self.frames.txtMaxLevel:GetText() or ""
    if minLevel ~= "" or maxLevel ~= "" then
        fields.lvl = ""
        if minLevel == maxLevel then
            fields.lvl = minLevel
        else
            if minLevel ~= "" then
                fields.lvl = ">" .. minLevel
            end
            if maxLevel ~= "" then
                fields.lvl = fields.lvl .. "<" .. maxLevel
            end
        end
    end
    end -- level
    
    do -- slot
    local slot = self.frames.txtSlot:GetText() or ""
    if slot ~= "" then
        fields.slot = slot
    end
    end -- slot

    do -- type
    local type = self.frames.txtType:GetText() or ""
    if type ~= "" then
        fields.type = type
    end
    end -- type
    
    do -- tooltip
    local tooltip = self.frames.txtTooltip:GetText() or ""
    if tooltip ~= "" then
        fields.tip = tooltip
    end
    end -- tooltip
    
    self:Search(fields)
end

-- Runs a search with the specified fields as seed parameters, and other 
-- parameters coming from the main window's search box
function ItemPreview:Search(fields)
    do -- Split search string
    
    -- search string should be in the format
    -- name tag:value tag:value
    local searchStr = self.frames.txtSearch:GetText() or ""
    searchFields = fields or {}
    
    -- See if the first field is the name (isn't colonated)
    local colonIndex = searchStr:find(":")
    if not colonIndex then
        searchFields.name = searchStr
        searchStr = ""
    else
        local lastSpaceIndex = Sol.string.LastIndexOf(searchStr:sub(0, colonIndex), " ")        
        if lastSpaceIndex then
            searchFields.name = searchStr:sub(0, lastSpaceIndex - 1)
            searchStr = searchStr:sub(lastSpaceIndex + 1)
        end
    end

    -- Get the rest of the fields
    local searchData = Sol.string.Split(searchStr, ":")
    local currentField = nil
    for i = 1, #searchData do
        local str = searchData[i]
        if currentField then
            if str:sub(1,1) == " " then
                str = str:sub(2)
            end
            searchFields[currentField] = str
            currentField = nil
        elseif str ~= "" then
            currentField = str
        end
    end
    end -- split search string

    if self.settings.IgnoreCase then
        for k, v in pairs(searchFields) do
            searchFields[k] = v:lower()
        end
    end
    
    self:UpdateList()
    self:ClearFocus()
end

-- Determine whether to show the specified item or not, based on the search fields
function ItemPreview:ShouldShowItem(item, quality)
    if not item then return false end
    
    do -- Check item quality
    
    local itemColor = item.color
    if quality == 1 or (self.constants.QUALITY_COLORS[quality] == itemColor) then
        -- 1 is any; the others are specific colors
        --return true
    elseif quality == #self.constants.QUALITY_COLORS then
        -- the last one is "All others"
        for i = 2, #self.constants.QUALITY_COLORS - 1 do
            if self.constants.QUALITY_COLORS[i] == itemColor then
                return false
            end
        end
        --return true
    else
        return false
    end
    end -- check quality
    
    
    do -- Check search text
    
    for type, searchStr in pairs(searchFields) do
        local searchFn = self.SearchFunctions[type]
        local negate = searchStr:match("^!")
        if negate then
            searchStr = searchStr:sub(2)
        end
        if searchStr and searchFn then
            local found = searchFn(self, searchStr, item, self.settings.IgnoreCase)
            if (found and negate) or (not found and not negate) then
                return false
            end
        end
    end
    end -- search
    
    return true
end

