local Sol = LibStub("Sol")


local lastUpdate  = nil    -- the last time the list frame was updated
local NUM_TO_SKIP = 300    -- the max number of items to skip for each id that doesn't have valid data
local NUM_PASSES  = 1
local tipLeftElements = {}
local tipRightElements = {}

-------------------------------------------------------------------------------
--  Refresh Functions                                                        --
-------------------------------------------------------------------------------

--[[
pass         -- the current refresh pass (0: none, 1: links, 2: tooltips, 3: data)
isRefreshing -- true if a refresh is going on, false if no refresh or one is paused
force        -- whether or not the current refresh is a forced one
startID      -- the starting id of the current refresh pass
endID        -- the final id of the current refresh pass
currentID    -- the next item id to check when refreshing
--]]
function ItemPreview:NewRefreshState(refreshPass, force)
    if refreshPass > NUM_PASSES then
        return nil
    end
    
    local state = {
        isRefreshing = false,
        force = force,
        pass = refreshPass,
        startID = self.constants.MIN_ID,
        endID = self.constants.MAX_ID,
        currentID = self.constants.MIN_ID,
        date = nil,
    }
    
    state.NextState = function(self)
        local retVal = ItemPreview:NewRefreshState(self.refreshPass + 1, self.force)
        retVal.isRefreshing = self.isRefreshing
        return retVal
    end
    state.TogglePaused = function(self)
        self.isRefreshing = not self.isRefreshing
    end
    state.CurrentItem = function(self)
        return ItemPreview.itemList[self.currentID]
    end
    state.NextID = function(self)
        self.currentID = self.currentID + 1        
        return self.currentID
    end
    
    return state
end

-- Adds or updates the count for the specified item's drops from the given zone
function ItemPreview:UpdateZoneForItem(id, zone)
    local item = self.itemList[id]
    if not item then 
        self:SaveCurrentLink(id, os.date("%x"))
        item = self.itemList[id]
    end
    if not item then return end
    
    if type(item.zones) ~= "table" then 
        item.zones = {}
    end
    if not item.zones[zone] then
        item.zones[zone] = 0
    end
    item.zones[zone] = item.zones[zone] + 1
end

-- Adds or updates the count for the specified item's drops from the given mob
function ItemPreview:UpdateMobForItem(id, mob)
    local item = self.itemList[id]
    if not item then 
        self:SaveCurrentLink(id, os.date("%x"))
        item = self.itemList[id]
    end
    if not item then return end
    
    if type(item.mobs) ~= "table" then
        item.mobs = {}
    end
    if not item.mobs[mob] then
        item.mobs[mob] = 0
    end
    item.mobs[mob] = item.mobs[mob] + 1
end

function ItemPreview:ResetListData()
    ItemPreview_ListData = { 
        count = 0, 
        minID = 99999999, 
        maxID = -1,
    }
    self.itemList = ItemPreview_ListData
end

-- Forces all data to be refreshed, even for those IDs that have data saved already
function ItemPreview:ForceRefresh()
    self:ResetListData()
    self.refreshState = nil
    self:RefreshList(true)
end

-- Begin, pause, or continue a refresh
function ItemPreview:RefreshList(force)
    self.startRefreshTime = GetTime()
	if not self.refreshState then 
        self.refreshState = self:NewRefreshState(1, force)
    end
    
    -- if this isn't the first refresh, pass an 'update' date with the state
    if self.itemList.originalRefreshDate then
        self.refreshState.date = os.date("%x")
    end
    
    if not self.refreshState.isRefreshing then
        self:ClearFocus()
		self.frames.btnRefresh:SetText("Pause")
	else
		self.frames.btnRefresh:SetText("Continue")
	end
    self.refreshState:TogglePaused()
end

-- Finalize refresh and reset fields
function ItemPreview:FinishRefresh()
    Sol.io.Print("ItemPreview: Finished import. Total import time: " .. math.floor(GetTime() - self.startRefreshTime) .. " seconds.")
    if not self.itemList.originalRefreshDate then
        self.itemList.originalRefreshDate = os.date("%x")
    end
    self.itemList.lastRefreshDate = self.refreshState.date
    self.refreshState = nil
    
    GameTooltip1:Hide()
    GameTooltip2:Hide()
    self.frames.lblTitle:SetText("ItemPreview (" .. self.itemList.count .. " itemsZ)")
    self.frames.btnRefresh:SetText("Refresh")
    self.frames.btnRefresh:Hide()
    
    local id = UIDropDownMenu_GetSelectedID(self.frames.ddRange)
    self.SelectRange(_, id, _)
    self.SelectSortMode(_, self.settings.SortMode, _)
    self.SelectRange(_, id, _)
end

-- Update the title text that displays refresh progress
function ItemPreview:UpdateListProgress()
    local percent = (self.refreshState.currentID - self.refreshState.startID) / (self.refreshState.endID - self.refreshState.startID)
    percent = math.ceil(percent * 100)
    self.frames.lblTitle:SetText(string.format("ItemPreview (Import: %d%%)", percent))
    
    -- give them a little something to tide them over ^^
    if self.frames.frmMain:IsVisible() and not lastUpdate or GetTime() - lastUpdate > self.settings.UpdateInterval then
        lastUpdate = GetTime()
        local id = UIDropDownMenu_GetSelectedID(self.frames.ddRange)
        self.SelectRange(_, id, _)
        self.SelectSortMode(_, self.settings.SortMode, _)
        self.SelectRange(_, id, _)
    end
end

function ItemPreview:InitTooltipLines()
    for i = 1, self.constants.TOOLTIP_LINES do
        tipLeftElements[i] = _G[self.frames.datatip:GetName() .. "TextLeft" .. i]
        tipRightElements[i] = _G[self.frames.datatip:GetName() .. "TextRight" .. i]
    end
end

-- Creates a link for the specified item id and saves it
-- Returns true if a tooltip was read at all
function ItemPreview:SaveCurrentLink(id, date, force)
    local item = self.itemList[id]
    if item and not force then return false end
    
    Sol.tooltip.ClearAllText(self.frames.datatip)
    local text, color = self:GetItemLinkDataFromID(string.format("%x", id))
    if text then
        local item = {
            id = id,
            text = text,
            color = color,
            date = date,
            show = nil,
        }
        if not self.itemList[id] then
            self.itemList.count = self.itemList.count + 1
        end
        self.itemList[id] = item
        
        if not self.itemList.minID or self.itemList.minID > id then
            self.itemList.minID = id
        end
        if not self.itemList.maxID or self.itemList.maxID < id then
            self.itemList.maxID = id
        end
        
        if self.settings.SaveTooltips and color ~= "ffffff" then
        --if self.settings.SaveTooltips then
            item.tip = { left = {}, right = {} }
            for i = 1, self.constants.TOOLTIP_LINES do
                local left = tipLeftElements[i]:GetText()
                local right = tipRightElements[i]:GetText()
                
                if left ~= "" then 
                    item.tip.left[i] = left
                    if i > 1 and i < 5 then
			if (not left:match(TEXT("C_PDAMAGE"))) and (not left:match(TEXT("C_PDEFENSE"))) and 
                           (not left:match(TEXT("C_MDEFENSE"))) and (not left:match(TEXT("C_MDAMAGE"))) and
                           (not left:lower():match(TEXT("_glossary_00338"))) and (left:match(TEXT("C_LEVEL"))) then
                               local lvl = left:match("^.*%s.*%s(%d+)$")
                               if lvl then item.lvl = tonumber(lvl) end
                        end
                    end
                    if right ~= "" then
                        item.tip.right[i] = right
                    end
                end
            end
        end
    end
    return true
end

-- Run a few iterations of the refresh loop (not enough to freeze the UI though)
function ItemPreview:RunRefreshLoop()
    local state = self.refreshState
    for i = 1, self.settings.UpdateSpeed do
        if state.currentID > state.endID then 
            self:FinishRefresh()
            return
        end
        
        if not self:SaveCurrentLink(state.currentID, state.date, state.force) then
            -- skip until an item is updated or it's time to move on
            for j = 1, NUM_TO_SKIP do
                state:NextID()
                if self:SaveCurrentLink(state.currentID, state.date, state.force) then
                    break
                end
            end
        end
        state:NextID()
    end
    
    self:UpdateListProgress()
end

